# pylint: disable=redefined-outer-name
from collections.abc import Iterator
from datetime import datetime
from logging import getLogger

import time_machine
from logikal_utils.project import tool_config
from pytest import Parser, fixture

FROZEN_TIMESTAMP = tool_config('pytest').get('frozen_time')
FROZEN_TIME = bool(FROZEN_TIMESTAMP)

time_machine.naive_mode = time_machine.NaiveMode.LOCAL

logger = getLogger(__name__)


def pytest_addoption(parser: Parser) -> None:
    parser.addini('frozen_time', 'the frozen timestamp to use')


@fixture(scope='session')
def frozen_timestamp() -> datetime:
    """
    Return the frozen timestamp.

    The frozen timestamp can be set under the ``frozen_time`` value in the ``tool.pytest`` section
    of ``pyproject.toml``. It must be in ``%Y-%m-%d %H:%M:%S`` format. When the frozen timestamp is
    not configured, it defaults to the current timestamp generated at session startup.
    """
    if FROZEN_TIMESTAMP:
        return datetime.strptime(FROZEN_TIMESTAMP, '%Y-%m-%d %H:%M:%S')
    return datetime.now()  # pragma: no cover, tested in subprocess


@fixture(autouse=FROZEN_TIME)
def frozen_time(
    frozen_timestamp: datetime,
) -> Iterator[time_machine.Traveller]:  # noqa: D400, D402, D415
    """
    frozen_time() -> time_machine.Traveller

    Freeze time and return a traveller instance for managing time.

    Automatically applied to each test when a frozen timestamp has been configured.
    """
    logger.debug(f'Setting the time to "{frozen_timestamp}"')
    with time_machine.travel(destination=frozen_timestamp, tick=False) as traveller:
        yield traveller



import inspect
from collections.abc import Iterable
from contextlib import AbstractContextManager, nullcontext
from dataclasses import replace
from functools import cache
from itertools import product
from logging import getLogger
from operator import itemgetter
from pathlib import Path
from typing import Any

import pytest
from logikal_browser import Browser
from logikal_browser.install import InstalledBrowser, installed_browsers
from logikal_browser.scenarios import Scenario, Settings

from pytest_logikal.core import EXTRAS
from pytest_logikal.utils import Fixture, Function

logger = getLogger(__name__)

__all__ = ['Browser', 'set_browser']

_INSTALLED_BROWSERS_KEY = pytest.StashKey[dict[str, InstalledBrowser]]()


def pytest_sessionstart(session: pytest.Session) -> None:
    session.config.stash[_INSTALLED_BROWSERS_KEY] = installed_browsers(
        install=not session.config.getoption('no_install'),
    )


def pytest_report_header(config: pytest.Config) -> str | list[str]:
    browsers = config.stash[_INSTALLED_BROWSERS_KEY].values()
    if config.get_verbosity() > 0:
        return ['browsers:'] + [f'  {repr(browser.browser_version)}' for browser in browsers]
    return f'browsers: {', '.join(str(browser.browser_version) for browser in browsers)}'


@cache
def _get_installed_browsers() -> dict[str, InstalledBrowser]:
    return installed_browsers()


def _get_scenarios(scenarios: Scenario | Iterable[Scenario]) -> Iterable[Scenario]:
    return [scenarios] if isinstance(scenarios, Scenario) else scenarios


def _get_settings(settings: Settings | Iterable[Settings]) -> Iterable[Settings]:
    return [settings] if isinstance(settings, Settings) else settings


def _get_languages(scenario_languages: Iterable[str] | None) -> Iterable[str] | Iterable[None]:
    # Use the scenario languages if Django is installed
    if scenario_languages:
        if not EXTRAS['django']:
            raise RuntimeError('The `django` extra must be installed for scenario languages')
        return scenario_languages

    # Use Django's configured languages if available
    if EXTRAS['django']:
        from django.conf import settings  # pylint: disable=import-outside-toplevel

        return [language_code for language_code, _ in settings.LANGUAGES]

    # Return `None` as a default
    return [None]


def _check_browser_fixture_request(request: Any) -> None:
    if (
        not hasattr(request, 'param')
        or 'installed_browser' not in request.param
        or 'settings' not in request.param
        or 'language' not in request.param
    ):
        raise RuntimeError('You must specify a browser scenario for the browser fixture')


def _language_context(language: str | None) -> AbstractContextManager[Any]:
    if not language:
        return nullcontext()

    if not EXTRAS['django']:
        raise RuntimeError('The `django` extra must be installed for browser languages')

    from django.utils import translation  # pylint: disable=import-outside-toplevel

    logger.debug(f'Setting the language to "{language}"')
    return translation.override(language)


@pytest.fixture
def browser(tmp_path: Path, request: Any) -> Iterable[Browser]:  # noqa: D400, D402, D415
    """
    browser() -> logikal_browser.Browser

    Yield a scenario-specific :class:`~logikal_browser.Browser` sub-class instance.
    """
    _check_browser_fixture_request(request)

    function = request.node.obj
    if not (source_file := inspect.getsourcefile(function)):  # pragma: no cover, defensive line
        raise RuntimeError(f'Source file cannot be found for "{function}"')
    source_path = Path(source_file)
    screenshot_path = source_path.parent / 'screenshots' / source_path.stem / function.__name__
    installed = request.param['installed_browser']
    language = request.param['language']
    with _language_context(language=language):
        with installed.browser_class(
            settings=request.param['settings'],
            version=installed.browser_version,
            language=language,
            screenshot_path=screenshot_path,
            screenshot_tmp_path=tmp_path,
        ) as driver:
            yield driver


def set_browser(
    scenarios: Scenario | Iterable[Scenario],
    languages: Iterable[str] | None = None,
    headless: bool | None = None,
) -> Fixture[Any]:
    """
    Apply the given scenarios to the :func:`~pytest_logikal.browser.browser` fixture.

    Args:
        scenarios: The scenarios to use.
        languages: The languages to use. Defaults to using the scenario languages.
        headless: Whether to use a headless browser instance.

    """
    def parametrized_test_function(function: Function) -> Any:
        argvalues = []
        installed = _get_installed_browsers()
        for scenario in _get_scenarios(scenarios):
            for settings, language, browser_name in product(
                _get_settings(scenario.settings),
                _get_languages(languages or scenario.languages),
                scenario.browsers or installed.keys(),
            ):
                name_parts = [settings.name, language, browser_name]
                argvalues.append({
                    'name': '-'.join(part for part in name_parts if part is not None),
                    'installed_browser': installed[browser_name],
                    'settings': (
                        settings if headless is None
                        else replace(settings, headless=headless)
                    ),
                    'language': language,
                })

        return pytest.mark.parametrize(
            argnames='browser', argvalues=argvalues, indirect=True, ids=itemgetter('name'),
        )(function)

    return parametrized_test_function



import zoneinfo
from collections.abc import Iterable, Sequence
from logging import getLogger
from typing import Any, Protocol

import pytest
from django.conf import settings
from django.urls import reverse
from django.utils import timezone as django_timezone, translation
from factory import random as factory_random
from logikal_utils.random import DEFAULT_RANDOM_SEED
from mypy_django_plugin import config as mypy_django_plugin_config
from pytest_django.live_server_helper import LiveServer

from pytest_logikal.node_install import install_node_packages
from pytest_logikal.utils import Fixture, Function
from pytest_logikal.validator import Validator

logger = getLogger(__name__)

__all__ = ['LiveURL', 'set_language', 'set_timezone']


def pytest_sessionstart(session: pytest.Session) -> None:
    if not session.config.getoption('no_install'):
        install_node_packages()

    css_or_svg = not session.config.getoption('no_css') or not session.config.getoption('no_svg')
    if not session.config.getoption('fast') and css_or_svg:
        # We start the validator service here to avoid pytest's output capturing
        Validator.service_url()


def pytest_configure(config: pytest.Config) -> None:
    # Patching django-stubs
    def parse_toml_file(self: Any, *_args: Any, **_kwargs: Any) -> None:
        self.django_settings_module = config.inicfg['DJANGO_SETTINGS_MODULE']
        self.strict_settings = True
        self.strict_model_abstract_attrs = True

    mypy_django_plugin_config.DjangoPluginConfig.parse_toml_file = parse_toml_file  # type: ignore


@pytest.fixture(scope='session', autouse=True)
def faker_seed() -> int:
    """
    Set a default seed for Faker for deterministic testing. Automatically applied.
    """
    return DEFAULT_RANDOM_SEED


@pytest.fixture(scope='function', autouse=True)
def factory_seed() -> None:
    """
    Set a default seed for pytest-factoryboy for deterministic testing. Automatically applied.
    """
    factory_random.reseed_random(DEFAULT_RANDOM_SEED)


class LiveURL(Protocol):
    def __call__(
        self,
        name: str | None = None,
        args: Sequence[Any] | None = None,
        kwargs: dict[str, Any] | None = None,
    ) -> str:
        ...


@pytest.fixture
def live_url(live_server: LiveServer) -> LiveURL:  # noqa: D400, D402, D415, D417
    """
    live_url(name: str | None, args: Sequence[Any] | None, kwargs: dict[str, Any] | None) -> str

    Return the path to a URL.

    Args:
        name: The URL pattern name.
        args: The URL arguments to use.
        kwargs: The URL keyword arguments to use.

    """
    def live_url_path(
        name: str | None = None,
        args: Sequence[Any] | None = None,
        kwargs: dict[str, Any] | None = None,
    ) -> str:
        return live_server.url + (reverse(viewname=name, args=args, kwargs=kwargs) if name else '')
    return live_url_path


@pytest.fixture
def language(request: Any) -> Iterable[str]:
    """
    Return the currently activated language.
    """
    if language_code := getattr(request, 'param', None):
        logger.debug(f'Setting the language to "{language_code}"')
        with translation.override(language_code):
            yield language_code
    else:
        yield translation.get_language()


def set_language(*language_codes: str) -> Fixture[Any]:
    """
    Mark a test to run with each of the specified languages.

    Args:
        *language_codes: The language codes to use.

    .. warning:: This decorator should not be used together with the :func:`browser
        <pytest_logikal.browser.browser>` fixture. Specify the language in the browser scenario or
        use the ``languages`` parameter of the :func:`set_browser
        <pytest_logikal.browser.set_browser>` decorator instead.

    .. note:: You must also use the :func:`language <pytest_logikal.django.language>` fixture in
        your test when applying this decorator.

    """
    def parametrized_test_function(function: Function) -> Any:
        if 'browser' in function.__code__.co_varnames:
            raise RuntimeError(
                'The `set_language` and `all_languages` decorators '
                'cannot be used together with the `browser` fixture'
            )
        return pytest.mark.parametrize(
            argnames='language', argvalues=language_codes, indirect=True,
        )(function)
    return parametrized_test_function


def all_languages() -> Fixture[Any]:
    """
    Mark a test to run with every available language.

    .. warning:: This decorator should not be used together with the :func:`browser
        <pytest_logikal.browser.browser>` fixture. When the ``django`` extra is installed, tests
        automatically run with all configured languages.

    .. note:: You must also use the :func:`language <pytest_logikal.django.language>` fixture in
        your test when applying this decorator.

    """
    return set_language(*(language_code for language_code, _ in settings.LANGUAGES))


@pytest.fixture
def timezone(request: Any) -> Iterable[str]:
    """
    Return the current time zone ID.
    """
    if zone_id := getattr(request, 'param', None):
        logger.debug(f'Setting the time zone to "{zone_id}"')
        current_timezone = django_timezone.get_current_timezone()
        django_timezone.activate(zoneinfo.ZoneInfo(zone_id))
        yield zone_id
        django_timezone.activate(current_timezone)
    else:
        yield django_timezone.get_current_timezone_name()


def set_timezone(*zone_ids: str) -> Fixture[Any]:
    """
    Mark a test to run with each of the specified time zones.

    Args:
        *zone_ids: The time zone IDs to use.

    .. note:: You must also use the :func:`timezone <pytest_logikal.django.timezone>` fixture in
        your test when applying this decorator.

    """
    def parametrized_test_function(function: Function) -> Any:
        return pytest.mark.parametrize(
            argnames='timezone', argvalues=zone_ids, indirect=True,
        )(function)
    return parametrized_test_function
